#include "core/List.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "core/Utility.h"

static void ensureCapacity(CoreList* l) {
    if(l->length >= l->capacity) {
        coreListReserve(l, l->capacity + coreMaxSize(4lu, l->capacity / 4));
    }
}

static void setSize(CoreList* l, size_t n) {
    l->capacity = n;
    l->data = coreReallocate(l->data, l->dataSize * n);
}

static void* getPointer(CoreList* l, size_t index) {
    return (char*)l->data + (l->dataSize * index);
}

static const void* getPointerC(const CoreList* l, size_t index) {
    return (const char*)l->data + (l->dataSize * index);
}

void coreCopyList(CoreList* l, const CoreList* other) {
    if(l == other) {
        return;
    }
    coreDestroyList(l);
    l->length = other->length;
    l->dataSize = other->dataSize;
    setSize(l, other->capacity);
    memcpy(l->data, other->data, l->dataSize * l->length);
}

void coreMoveList(CoreList* l, CoreList* other) {
    if(l == other) {
        return;
    }
    coreSwapList(l, other);
    coreDestroyList(other);
}

void coreDestroyList(CoreList* l) {
    coreFree(l->data);
    *l = CORE_LIST(0);
}

void coreListReserve(CoreList* l, size_t n) {
    if(n > l->capacity) {
        setSize(l, n);
    }
}

void coreShrinkList(CoreList* l) {
    if(l->length != l->capacity) {
        setSize(l, l->length);
    }
}

void coreResizeList(CoreList* l, size_t n) {
    if(l->length < n) {
        coreListReserve(l, n);
        memset(getPointer(l, l->length), 0, (n - l->length) * l->dataSize);
    }
    l->length = n;
}

void coreResizeListPointer(CoreList* l, size_t n, const void* data) {
    if(l->length < n) {
        coreListReserve(l, n);
        while(l->length < n) {
            coreListAddPointer(l, data);
        }
    } else if(l->length > n) {
        l->length = n;
    }
}

CoreList* coreListAddPointer(CoreList* l, const void* data) {
    ensureCapacity(l);
    memcpy(getPointer(l, l->length++), data, l->dataSize);
    return l;
}

CoreList* coreListAddLast(CoreList* l) {
    return coreListAddPointer(l, coreListLastVoidPointer(l));
}

void* coreListGetVoidPointer(CoreList* l, size_t index) {
    assert(index < l->length);
    return getPointer(l, index);
}

const void* coreListGetVoidPointerC(const CoreList* l, size_t index) {
    assert(index < l->length);
    return getPointerC(l, index);
}

void* coreListLastVoidPointer(CoreList* l) {
    return coreListGetVoidPointer(l, l->length - 1);
}

const void* coreListLastVoidPointerC(const CoreList* l) {
    return coreListGetVoidPointerC(l, l->length - 1);
}

void coreClearList(CoreList* l) {
    l->length = 0;
}

void coreListRemoveBySwap(CoreList* l, size_t index) {
    assert(index < l->length);
    size_t length = l->length - 1;
    if(index != length) {
        memcpy(getPointer(l, index), getPointer(l, length), l->dataSize);
    }
    l->length = length;
}

void coreListRemove(CoreList* l, size_t index) {
    assert(index < l->length);
    l->length--;
    char* p = getPointer(l, index);
    char* end = getPointer(l, l->length);
    while(p != end) {
        *p = *(p + l->dataSize);
        p++;
    }
}

void coreListRemoveLast(CoreList* l) {
    coreListRemoveBySwap(l, l->length - 1);
}

size_t coreToStringList(CoreList* l, char* buffer, size_t n, CoreToString c) {
    size_t w = 0;
    coreStringAddI(&w, &buffer, &n, snprintf(buffer, n, "["));
    char* end = getPointer(l, l->length);
    char* p = l->data;
    while(p != end) {
        coreStringAdd(&w, &buffer, &n, c(p, buffer, n));
        p += l->dataSize;
        if(p != end) {
            coreStringAddI(&w, &buffer, &n, snprintf(buffer, n, ", "));
        }
    }
    coreStringAddI(&w, &buffer, &n, snprintf(buffer, n, "]"));
    return w;
}

void coreSwapList(CoreList* a, CoreList* b) {
    CoreList tmp = *a;
    *a = *b;
    *b = tmp;
}
